from argparse import ArgumentParser, Namespace, _SubParsersAction
from datetime import datetime

from tabulate import tabulate

from pipeline.cloud import http
from pipeline.cloud.compute_requirements import Accelerator
from pipeline.cloud.schemas import pipelines as pipelines_schema
from pipeline.cloud.schemas.pagination import (
    Paginated,
    get_default_pagination,
    to_page_position,
)
from pipeline.util.logging import _print


def edit_parser(command_parser: "_SubParsersAction[ArgumentParser]") -> None:
    edit_parser = command_parser.add_parser(
        "pipelines",
        aliases=["pipeline", "pl"],
        help="Edit a pipeline.",
    )

    edit_parser.set_defaults(func=_edit_pipeline)

    edit_parser.add_argument(
        "pipeline_id",
        help="Pipeline to edit.",
    )

    edit_parser.add_argument(
        "--cache-number",
        "-c",
        help="Minimum cache number.",
        type=int,
    )

    edit_parser.add_argument(
        "--gpu-memory",
        "-g",
        help="Minimum GPU memory.",
        type=int,
    )


def get_parser(command_parser: "_SubParsersAction[ArgumentParser]") -> None:
    get_parser = command_parser.add_parser(
        "pipelines",
        aliases=["pipeline", "pl"],
        help="Get pipeline information.",
    )

    get_parser.set_defaults(func=_get_pipeline)

    # get by name
    # get_parser.add_argument(
    #     "--name",
    #     "-n",
    #     help="Pipeline name.",
    #     type=str,
    # )
    get_parser.add_argument(
        "--skip",
        "-s",
        help="Number of pipelines to skip in paginated set.",
        type=int,
    )
    get_parser.add_argument(
        "--limit",
        "-l",
        help="Total number of pipelines to fetch in paginated set.",
        type=int,
    )


def delete_parser(command_parser: "_SubParsersAction[ArgumentParser]") -> None:
    delete_parser = command_parser.add_parser(
        "pipelines",
        aliases=["pipeline", "pl"],
        help="Delete a pipeline.",
    )

    delete_parser.set_defaults(func=_delete_pipeline)

    delete_parser.add_argument(
        "pipeline_id",
        help="Pipeline to delete.",
    )


def _get_pipeline(args: Namespace) -> None:
    _print("Getting pipelines")

    # cluster_api = PipelineCloud(verbose=False)

    params = dict()
    pagination = get_default_pagination()
    # if name := getattr(args, "name", None):
    #     params["name"] = name
    if skip := getattr(args, "skip", None):
        pagination.skip = skip
    if limit := getattr(args, "limit", None):
        pagination.limit = limit
    paginated_raw_pipelines: Paginated[dict] = http.get(
        "/v4/pipelines",
        params=dict(**params, **pagination.dict()),
    ).json()

    # pipelines_: Paginated[pipelines_schema.PipelineGet] = Paginated[
    #     pipelines_schema.PipelineGet
    # ].parse_obj(paginated_raw_pipelines)

    pipelines = [
        [
            pipeline_raw["id"],
            pipeline_raw["name"],
            datetime.fromtimestamp(pipeline_raw.get("created_at"))
            if "created_at" in pipeline_raw
            else "N/A",
            val if (val := pipeline_raw.get("minimum_cache_number", "N/A")) else "N/A",
            (
                ""
                if not (accelerators := pipeline_raw.get("accelerators", None))
                else (
                    "nvidia_all"
                    if Accelerator.nvidia_all in accelerators
                    else (
                        "cpu"
                        if Accelerator.cpu in pipeline_raw.get("accelerators", [])
                        else "\n".join(
                            [
                                f"{accelerators.count(accelerator)}× {accelerator}"
                                for accelerator in set(accelerators)
                            ]
                        )
                    )
                )
            ),
        ]
        for pipeline_raw in paginated_raw_pipelines["data"]
    ]

    page_position = to_page_position(
        paginated_raw_pipelines["skip"],
        paginated_raw_pipelines["limit"],
        paginated_raw_pipelines["total"],
    )

    table = tabulate(
        pipelines,
        headers=[
            "ID",
            "Name",
            "Created",
            "Cache # (min-max)",
            "Accelerators",
        ],
        tablefmt="psql",
    )
    print(table)
    print(f"\nPage {page_position['current']} of {page_position['total']}\n")


def _edit_pipeline(args: Namespace) -> None:
    pipeline_id = getattr(args, "pipeline_id")
    cache_number = getattr(args, "cache_number", None)
    gpu_memory = getattr(args, "gpu_memory", None)

    patch_schema = pipelines_schema.PipelinePatch(
        minimum_cache_number=cache_number,
        gpu_memory_min=gpu_memory,
    )

    if cache_number is None and gpu_memory is None:
        _print("Nothing to edit.", level="ERROR")
        return

    http.patch(
        f"/v4/pipelines/{pipeline_id}",
        patch_schema.dict(),
    )

    _print("Pipeline edited!")


def _delete_pipeline(args: Namespace) -> None:
    pipeline_id = getattr(args, "pipeline_id")

    http.delete(
        f"/v4/pipelines/{pipeline_id}",
    )

    _print("Pipeline deleted!")



from collections import defaultdict
from enum import Enum


class Accelerator(str, Enum):
    nvidia_t4: str = "nvidia_t4"
    nvidia_a100: str = "nvidia_a100"
    nvidia_a100_80gb: str = "nvidia_a100_80gb"
    nvidia_v100: str = "nvidia_v100"
    nvidia_v100_32gb: str = "nvidia_v100_32gb"
    nvidia_3090: str = "nvidia_3090"
    nvidia_a16: str = "nvidia_a16"
    nvidia_h100: str = "nvidia_h100"
    nvidia_l4: str = "nvidia_l4"
    nvidia_a5000: str = "nvidia_a5000"
    nvidia_all: str = "nvidia_all"
    cpu: str = "cpu"

    @classmethod
    def from_str(cls, accelerator: str) -> "Accelerator":
        if "T4" in accelerator:
            accelerator_type = Accelerator.nvidia_t4
        elif "A100" in accelerator:
            if "80GB" in accelerator:
                accelerator_type = Accelerator.nvidia_a100_80gb
            else:
                accelerator_type = Accelerator.nvidia_a100
        elif "H100" in accelerator:
            accelerator_type = Accelerator.nvidia_h100
        elif "V100" in accelerator:
            if "32GB" in accelerator:
                accelerator_type = Accelerator.nvidia_v100_32gb
            else:
                accelerator_type = Accelerator.nvidia_v100
        elif "3090" in accelerator:
            accelerator_type = Accelerator.nvidia_3090
        elif "A16" in accelerator:
            accelerator_type = Accelerator.nvidia_a16
        elif "L4" in accelerator:
            accelerator_type = Accelerator.nvidia_l4
        elif "A5000" in accelerator:
            accelerator_type = Accelerator.nvidia_a5000
        else:
            raise Exception(f"Unknown GPU name: {accelerator}")

        return accelerator_type

    @classmethod
    def valid_accelerator_config(cls, accelerators: list["Accelerator"]):
        return accelerators in [
            [Accelerator.nvidia_t4],
            [Accelerator.nvidia_a100],
            [Accelerator.nvidia_a100] * 2,
            [Accelerator.nvidia_a100] * 4,
            [Accelerator.nvidia_a100_80gb],
            [Accelerator.nvidia_a100_80gb] * 2,
            [Accelerator.nvidia_a100_80gb] * 4,
            [Accelerator.nvidia_v100],
            [Accelerator.nvidia_v100_32gb],
            [Accelerator.nvidia_l4],
            [Accelerator.nvidia_a5000],
            [Accelerator.nvidia_all],
            [Accelerator.cpu],
        ]

    def max_memory_mb(self) -> int:
        return defaultdict(
            lambda: 10_000,
            {
                Accelerator.nvidia_t4: 16_000,
                Accelerator.nvidia_a100: 40_000,
                Accelerator.nvidia_a100_80gb: 80_000,
                Accelerator.nvidia_v100: 16_000,
                Accelerator.nvidia_v100_32gb: 32_000,
                Accelerator.nvidia_l4: 22_000,
                Accelerator.nvidia_a5000: 24_000,
            },
        )[self]


nvidia_gpus = [
    Accelerator.nvidia_t4,
    Accelerator.nvidia_a100,
    Accelerator.nvidia_a100_80gb,
    Accelerator.nvidia_v100,
    Accelerator.nvidia_v100_32gb,
    Accelerator.nvidia_3090,
    Accelerator.nvidia_a16,
    Accelerator.nvidia_h100,
    Accelerator.nvidia_l4,
    Accelerator.nvidia_a5000,
    Accelerator.nvidia_all,
]



import asyncio
import io
import os
import shutil
import uuid
from pathlib import Path

from fastapi import APIRouter, Request, Response
from loguru import logger

from pipeline.cloud.schemas import pipelines as pipeline_schemas
from pipeline.cloud.schemas import runs as run_schemas
from pipeline.container.manager import Manager
from pipeline.exceptions import RunInputException, RunnableError
from pipeline.objects.graph import File

router = APIRouter(prefix="/runs")


@router.post(
    "",
    tags=["runs"],
    status_code=200,
    # response_model=run_schemas.Run,
    responses={
        500: {
            "description": "Pipeline failed",
            "model": run_schemas.ContainerRunResult,
        },
        400: {
            "description": "Invalid input data",
            "model": run_schemas.ContainerRunResult,
        },
    },
)
async def run(
    run_create: run_schemas.ContainerRunCreate,
    request: Request,
    response: Response,
):
    run_id = run_create.run_id
    with logger.contextualize(run_id=run_id):
        manager: Manager = request.app.state.manager
        if manager.pipeline_state == pipeline_schemas.PipelineState.loading:
            logger.info("Pipeline loading")
            return run_schemas.ContainerRunResult(
                outputs=None,
                error=run_schemas.ContainerRunError(
                    type=run_schemas.ContainerRunErrorType.pipeline_loading,
                    message="Pipeline is still loading",
                ),
            )

        if manager.pipeline_state == pipeline_schemas.PipelineState.failed:
            logger.info("Pipeline failed to load")
            return run_schemas.ContainerRunResult(
                outputs=None,
                error=run_schemas.ContainerRunError(
                    type=run_schemas.ContainerRunErrorType.startup_error,
                    message="Pipeline failed to load",
                    traceback=manager.pipeline_state_message,
                ),
            )

        execution_queue: asyncio.Queue = request.app.state.execution_queue

        response_queue: asyncio.Queue = asyncio.Queue()
        execution_queue.put_nowait((run_create, response_queue))
        run_output = await response_queue.get()

        response_schema, response.status_code = _generate_run_result(run_output)
        return response_schema


def _generate_run_result(run_output) -> tuple[run_schemas.ContainerRunResult, int]:
    if isinstance(run_output, RunInputException):
        return (
            run_schemas.ContainerRunResult(
                outputs=None,
                error=run_schemas.ContainerRunError(
                    type=run_schemas.ContainerRunErrorType.input_error,
                    message=run_output.message,
                ),
            ),
            400,
        )
    elif isinstance(run_output, RunnableError):
        return (
            run_schemas.ContainerRunResult(
                outputs=None,
                error=run_schemas.ContainerRunError(
                    type=run_schemas.ContainerRunErrorType.pipeline_error,
                    message=repr(run_output.exception),
                    traceback=run_output.traceback,
                ),
            ),
            200,
        )
    elif isinstance(run_output, Exception):
        return (
            run_schemas.ContainerRunResult(
                outputs=None,
                error=run_schemas.ContainerRunError(
                    type=run_schemas.ContainerRunErrorType.unknown,
                    message=str(run_output),
                ),
            ),
            500,
        )
    else:
        outputs = _parse_run_outputs(run_output)
        return (
            run_schemas.ContainerRunResult(
                outputs=outputs,
                error=None,
            ),
            200,
        )


def _parse_run_outputs(run_outputs):
    outputs = []
    for output in run_outputs:
        output_type = run_schemas.RunIOType.from_object(output)
        # if single file
        if output_type == run_schemas.RunIOType.file:
            file_schema = _save_run_file(output)
            outputs.append(
                run_schemas.RunOutput(type=output_type, value=None, file=file_schema)
            )
        # else if list of files
        elif (
            output_type == run_schemas.RunIOType.pkl
            and isinstance(output, list)
            and all([isinstance(item, (File, io.BufferedIOBase)) for item in output])
        ):
            file_list = []
            for file in output:
                file_schema = _save_run_file(file)
                file_list.append(
                    run_schemas.RunOutput(
                        type=run_schemas.RunIOType.file,
                        value=None,
                        file=file_schema,
                    )
                )
            outputs.append(
                run_schemas.RunOutput(
                    type=run_schemas.RunIOType.array,
                    value=file_list,
                    file=None,
                )
            )
        else:
            outputs.append(
                run_schemas.RunOutput(type=output_type, value=output, file=None)
            )
    return outputs


def _save_run_file(file: File | io.BufferedIOBase) -> run_schemas.RunOutputFile:
    # ensure we save the file somewhere unique on the file system so it doesn't
    # get overwritten by another run
    uuid_path = str(uuid.uuid4())
    output_path = Path(f"/tmp/run_files/{uuid_path}")
    output_path.mkdir(parents=True, exist_ok=True)

    if isinstance(file, File):
        # should always exist
        assert file.path
        file_name = file.path.name
        output_file = output_path / file_name
        file_size = os.stat(file.path).st_size
        # copy file to new output location
        shutil.copyfile(file.path, output_file)
    else:
        file_name = getattr(
            file,
            "name",
            str(uuid.uuid4()),
        )
        try:
            file_size = file.seek(0, os.SEEK_END)
        except Exception:
            file_size = -1
            logger.warning(f"Could not get size of type {type(file)}")
        # write file to new output location
        output_file = output_path / file_name
        output_file.write_bytes(file.read())

    file_schema = run_schemas.RunOutputFile(
        name=file_name, path=str(output_file), size=file_size, url=None
    )
    return file_schema



import json
import re
from argparse import ArgumentParser, Namespace, _SubParsersAction

from tabulate import tabulate

from pipeline.cloud import http
from pipeline.cloud.schemas import pointers as pointers_schema
from pipeline.cloud.schemas.pagination import (
    Paginated,
    get_default_pagination,
    to_page_position,
)
from pipeline.util.logging import _print

VALID_TAG_NAME = re.compile(
    r"^[a-z0-9][a-z0-9-._/]*[a-z0-9]:[0-9A-Za-z_][0-9A-Za-z-_.]{0,127}$"
)


def create_parser(command_parser: "_SubParsersAction[ArgumentParser]") -> None:
    create_parser = command_parser.add_parser(
        "pointers",
        aliases=["pointer", "ptr"],
        help="Create a new pointer.",
    )

    create_parser.add_argument(
        "new_pointer",
        type=str,
        help="Pointer to create.",
    )

    create_parser.add_argument(
        "pipeline_id_or_pointer",
        type=str,
        help="Pipeline id or pointer to create a pointer to.",
    )

    create_parser.add_argument(
        "--locked",
        action="store_true",
        help="Lock the pointer.",
    )

    create_parser.set_defaults(func=_create_pointer)


def edit_parser(command_parser: "_SubParsersAction[ArgumentParser]") -> None:
    edit_parser = command_parser.add_parser(
        "pointers",
        aliases=["pointer", "ptr"],
        help="Edit a pointer.",
    )

    edit_parser.add_argument(
        "pointer",
        type=str,
        help="Pointer to edit.",
    )

    edit_parser.add_argument(
        "--source",
        "-s",
        type=str,
        help="Pipeline id or pointer to create a pointer to.",
        default=None,
    )

    edit_parser.add_argument(
        "--locked",
        action="store_true",
        help="Lock the pointer.",
    )
    edit_parser.add_argument(
        "--unlocked",
        action="store_true",
        help="Lock the pointer.",
    )
    edit_parser.set_defaults(func=_edit_pointer)


def get_parser(command_parser: "_SubParsersAction[ArgumentParser]") -> None:
    get_parser = command_parser.add_parser(
        "pointers",
        aliases=["pointer", "ptr"],
        help="Get pointer information.",
    )

    get_parser.set_defaults(func=_get_pointer)

    get_parser.add_argument(
        "--name",
        "-n",
        type=str,
        help="Pipeline name.",
    )

    get_parser.add_argument(
        "--show-deleted",
        "-d",
        action="store_true",
        help="Show pointers to deleted pipelines.",
    )
    get_parser.add_argument(
        "--skip",
        "-s",
        help="Number of pointers to skip in paginated set.",
        type=int,
    )
    get_parser.add_argument(
        "--limit",
        "-l",
        help="Total number of pointer to fetch in paginated set.",
        type=int,
    )


def delete_parser(command_parser: "_SubParsersAction[ArgumentParser]") -> None:
    delete_parser = command_parser.add_parser(
        "pointers",
        aliases=["pointer", "ptr"],
        help="Delete a pointer.",
    )

    delete_parser.add_argument(
        "pointer",
        type=str,
        help="Pointer to delete.",
    )

    delete_parser.set_defaults(func=_delete_pointer)


def _get_pointer(namespace: Namespace) -> None:
    _print("Getting pointers")

    pipeline_name = getattr(namespace, "name", None)
    show_deleted = getattr(namespace, "show_deleted", False)
    query_params = dict()
    pagination = get_default_pagination()
    if skip := getattr(namespace, "skip", None):
        pagination.skip = skip
    if limit := getattr(namespace, "limit", None):
        pagination.limit = limit

    query_params["show_deleted"] = show_deleted
    if pipeline_name:
        query_params["pipeline_name"] = pipeline_name
    paginated_raw_pointers: Paginated[dict] = http.get(
        "/v4/pointers", params=dict(**query_params, **pagination.dict())
    ).json()

    pointers = [
        [
            pointer_raw["pointer"],
            pointer_raw["pipeline_id"],
        ]
        for pointer_raw in paginated_raw_pointers["data"]
    ]
    page_position = to_page_position(
        paginated_raw_pointers["skip"],
        paginated_raw_pointers["limit"],
        paginated_raw_pointers["total"],
    )

    table = tabulate(
        pointers,
        headers=["Pointer", "Pipeline ID"],
        tablefmt="psql",
    )
    print(table)
    print(f"\nPage {page_position['current']} of {page_position['total']}\n")


def _create_pointer(namespace: Namespace) -> None:
    new_pointer = getattr(namespace, "new_pointer")
    pipeline_id_or_pointer = getattr(namespace, "pipeline_id_or_pointer")
    locked = getattr(namespace, "locked", False)

    create_schema = pointers_schema.PointerCreate(
        pointer=new_pointer,
        pointer_or_pipeline_id=pipeline_id_or_pointer,
        locked=locked,
    )
    result = http.post(
        "/v4/pointers",
        json.loads(
            create_schema.json(),
        ),
    )

    pointer = pointers_schema.PointerGet.parse_obj(result.json())

    _print(f"Created pointer {pointer.pointer} -> {pointer.pipeline_id}")


def _edit_pointer(namespace: Namespace) -> None:
    pointer = getattr(namespace, "pointer")
    source = getattr(namespace, "source", None)
    locked = getattr(namespace, "locked", None)
    unlocked = getattr(namespace, "unlocked", None)

    if locked and unlocked:
        _print("Cannot lock and unlock at the same time!", level="ERROR")
        return

    if source is None and not locked and not unlocked:
        _print("Nothing to edit!", level="ERROR")
        return

    _print(f"Editing pointer ({pointer})")

    edit_schema = pointers_schema.PointerPatch(
        pointer_or_pipeline_id=source,
        locked=locked if locked is not None else not unlocked,
    )
    http.patch(
        f"/v4/pointers/{pointer}",
        json.loads(
            edit_schema.json(),
        ),
    )

    _print("Pointer edited!")


def _delete_pointer(namespace: Namespace) -> None:
    pointer = getattr(namespace, "pointer")

    _print(f"Deleting pointer ({pointer})")

    http.delete(
        f"/v4/pointers/{pointer}",
    )

    _print("Pointer deleted!")



from fastapi import APIRouter

from pipeline.container.routes.v4.container import router as v4_container_router
from pipeline.container.routes.v4.files import router as v4_file_router
from pipeline.container.routes.v4.runs import router as v4_runs_router

router = APIRouter(prefix="/v4")

router.include_router(v4_runs_router)
router.include_router(v4_file_router)
router.include_router(v4_container_router)



import asyncio
from pathlib import Path
from uuid import uuid4

import httpx
from fastapi import APIRouter, HTTPException, UploadFile, status
from fastapi.responses import FileResponse
from loguru import logger

from pipeline.cloud.schemas import files as files_schemas

router = APIRouter(prefix="/files", tags=["Files"])


@router.get("/download/{path:path}", status_code=status.HTTP_200_OK)
async def read_file(path: str):
    """Download the contents of a file stored on the container."""
    file_path = Path(path)
    if not file_path.exists() or not file_path.is_file():
        raise HTTPException(status_code=404, detail="File not found")
    return FileResponse(path)


@router.post(
    "",
    status_code=status.HTTP_201_CREATED,
)
async def file_upload(
    pfile: UploadFile,
):
    """Upload a file to the container. Returns the path of the file on the container."""
    data = await pfile.read()
    pfile_uid = str(uuid4())
    name = getattr(pfile, "filename")
    name = Path(name).name

    logger.info(f"Received file: {name}")
    pfile_path = Path(f"/tmp/{pfile_uid[:2]}/{pfile_uid[2:4]}/{name}")
    pfile_path.parent.mkdir(parents=True, exist_ok=True)

    with open(pfile_path, "wb") as f:
        f.write(data)

    return dict(path=str(pfile_path))


@router.post(
    "/upload-to-storage",
    status_code=status.HTTP_201_CREATED,
)
async def upload_local_files_to_storage(
    payload: files_schemas.UploadFilesToRemoteStorageCreate,
):
    """For list of files in the request, upload the file from local storage to
    remote storage, using presigned URL. Files are uploaded in parallel for
    efficiency.
    """
    async with httpx.AsyncClient() as client:
        await asyncio.gather(
            *[_upload_file_using_presigned_url(client, file) for file in payload.files]
        )


async def _upload_file_using_presigned_url(
    client: httpx.AsyncClient,
    file: files_schemas.UploadFileUsingPresignedUrl,
):
    logger.info(f"Uploading {file.local_file_path} to {file.upload_url}")
    with open(file.local_file_path, "rb") as f:
        response = await client.post(
            file.upload_url, files={"file": f}, data=file.upload_fields
        )
        if response.status_code >= 300:
            raise Exception(
                f"Error uploading file using presigned URL: {response.text}"
            )



import argparse
import sys
import traceback
from typing import List, Optional

from pipeline.console.commands import (
    cluster_parser,
    container_parser,
    create_parser,
    delete_parser,
    edit_parser,
    get_parser,
    logs_parser,
)


def construct_cli() -> argparse.ArgumentParser:
    base_parser = argparse.ArgumentParser(
        prog="pipeline",
        description="Create or run pipelines locally or in the cloud!",
        add_help=True,
    )

    base_parser.add_argument(
        "-v",
        "--verbose",
        help="Verbose logging",
        action="store_true",
    )

    command_parser = base_parser.add_subparsers(dest="command")

    create_parser(command_parser)
    get_parser(command_parser)
    edit_parser(command_parser)
    delete_parser(command_parser)
    cluster_parser(command_parser)
    logs_parser(command_parser)
    container_parser(command_parser)

    return base_parser


def execute_cli(
    parser: argparse.ArgumentParser,
    args: Optional[List[str]] = None,
) -> None:
    if args is None:
        args = sys.argv[1:]

    parsed_args = parser.parse_args(args=args)

    if (selected_func := getattr(parsed_args, "func", None)) is not None:
        selected_func(parsed_args)

    if parsed_args.command is None:
        parser.print_help()
        return


def _run() -> int:
    try:
        parser = construct_cli()
        execute_cli(parser)
    except Exception:
        traceback.print_exc()
        sys.exit(1)

    sys.exit(0)



import pkg_resources
from fastapi import APIRouter
from fastapi.responses import HTMLResponse

router = APIRouter(prefix="/play", tags=["play"])


@router.get("", response_class=HTMLResponse)
async def render_pipeline_play():
    ts_code = pkg_resources.resource_string(
        "pipeline", "container/frontend/index.html"
    ).decode("utf-8")

    return f"""{ts_code}"""



from pipeline.objects.decorators import entity, pipe
from pipeline.objects.function import Function
from pipeline.objects.graph import Directory, File, Graph, Variable
from pipeline.objects.model import Model
from pipeline.objects.pipeline import Pipeline
from pipeline.objects.wrappers import onnx_to_pipeline

__all__ = [
    "Pipeline",
    "Graph",
    "Function",
    "Model",
    "Variable",
    "pipe",
    "entity",
    "Directory",
    "File",
    "onnx_to_pipeline",
]



import typing as t

from fastapi import APIRouter, Request, Response

from pipeline.cloud.schemas import pipelines as pipeline_schemas
from pipeline.container.manager import Manager
from pipeline.objects import Graph

router = APIRouter(prefix="/container")


@router.get(
    "/state",
    tags=["container"],
    status_code=200,
    response_model=pipeline_schemas.PipelineContainerState,
    responses={
        503: {
            "description": "Pipeline is loading",
            "model": pipeline_schemas.PipelineContainerState,
        },
        500: {
            "description": "Pipeline failed to load",
            "model": pipeline_schemas.PipelineContainerState,
        },
    },
)
async def is_ready(request: Request, response: Response):
    run_manager = request.app.state.manager
    if run_manager.pipeline_state in [
        pipeline_schemas.PipelineState.loading,
        pipeline_schemas.PipelineState.not_loaded,
    ]:
        response.status_code = 503
    elif run_manager.pipeline_state == pipeline_schemas.PipelineState.failed:
        response.status_code = 500

    return pipeline_schemas.PipelineContainerState(
        state=run_manager.pipeline_state,
        message=run_manager.pipeline_state_message,
    )


@router.get(
    "/pipeline",
    tags=["container"],
    response_model=pipeline_schemas.Pipeline,
)
async def get_pipeline(request: Request):
    run_manager: Manager = request.app.state.manager
    input_variables: t.List[pipeline_schemas.IOVariable] = []
    output_variables: t.List[pipeline_schemas.IOVariable] = []
    graph: Graph = run_manager.pipeline
    for variable in graph.variables:
        if variable.is_input:
            input_variables.append(variable.to_io_schema())

        if variable.is_output:
            output_variables.append(variable.to_io_schema())

    input_variables = input_variables
    output_variables = output_variables
    return pipeline_schemas.Pipeline(
        name=run_manager.pipeline_name,
        image=run_manager.pipeline_image,
        input_variables=input_variables,
        output_variables=output_variables,
    )



import os
from typing import List

from pip._internal.commands.freeze import freeze


class Environment:
    def __init__(
        self,
        environment_name: str = None,
        dependencies: List[str] = [],
    ):
        self.environment_name = environment_name
        self.dependencies = dependencies

    def to_requirements(self, output_dir: str = "./") -> None:
        requirements_path = os.path.join(os.path.join(output_dir, "requirements.txt"))
        with open(requirements_path, "w") as req_file:
            for _dep in self.dependencies:
                req_file.write(f"{_dep}\n")

    def add_dependency(self, dependency: str) -> None:
        self.dependencies.append(dependency)

    @classmethod
    def from_requirements(cls, requirements_path: str, environment_name: str = None):
        if not os.path.exists(requirements_path):
            raise FileNotFoundError(
                f"Could not find the requirements file '{requirements_path}'"
            )

        with open(requirements_path, "r") as req_file:
            requirements_str_list = req_file.readlines()

        requirements_list = [
            _req.trim() for _req in requirements_str_list if not _req.startswith("#")
        ]
        return cls(environment_name=environment_name, dependencies=requirements_list)

    @classmethod
    def from_current(cls, environment_name: str = None):
        deps = [dep for dep in freeze.freeze() if dep.split() > 0]
        return cls(environment_name=environment_name, dependencies=deps)



import io
import json
import typing as t
from datetime import datetime
from enum import Enum

from pipeline.cloud.schemas import BaseModel


class RunState(str, Enum):
    created = "created"
    routing = "routing"
    queued = "queued"
    running = "running"
    completed = "completed"
    failed = "failed"
    no_resources_available = "no_resources_available"
    unknown = "unknown"

    @staticmethod
    def is_terminal(state: "RunState") -> bool:
        return state in RunState.terminal_states()

    @classmethod
    def terminal_states(cls) -> list["RunState"]:
        return [
            RunState.completed,
            RunState.failed,
            RunState.no_resources_available,
        ]

    @classmethod
    def __get_validators__(cls):
        cls.lookup = {v: k.value for v, k in cls.__members__.items()}
        cls.value_lookup = {k.value: v for v, k in cls.__members__.items()}
        yield cls.validate

    @classmethod
    def validate(cls, v):
        try:
            v = int(v)
        except Exception:
            ...

        if isinstance(v, str):
            try:
                state = cls.lookup[v]
            except KeyError:
                state = cls.unknown
            return state
        elif isinstance(v, str):
            try:
                state = getattr(cls, cls.value_lookup[v])
            except KeyError:
                state = cls.unknown
            return state
        else:
            raise ValueError(f"Invalid value: {v}")


class RunErrorType(Enum):
    input_error = 1
    unroutable = 2
    graph_error = 3
    runtime_error = 4


class RunError(BaseModel):
    exception: str
    traceback: t.Optional[str]


class RunFileType(Enum):
    input = "input"
    output = "output"


class RunFile(BaseModel):
    id: str
    run_id: str
    io_type: RunFileType
    path: str

    class Config:
        # use_enum_values = True
        orm_mode = True


class RunIOType(str, Enum):
    integer = "integer"
    string = "string"
    fp = "fp"
    dictionary = "dictionary"
    boolean = "boolean"
    none = "none"
    array = "array"
    pkl = "pkl"
    file = "file"

    @classmethod
    def from_object(cls, obj: t.Any):
        # Get the enum type for the object.
        from pipeline.objects.graph import File

        if isinstance(obj, int) or obj is int:
            return cls.integer
        elif isinstance(obj, float) or obj is float:
            return cls.fp
        elif isinstance(obj, str) or obj is str:
            return cls.string
        elif isinstance(obj, bool) or obj is bool:
            return cls.boolean
        elif obj is None:
            return cls.none
        elif isinstance(obj, dict) or obj is dict:
            if obj is dict:
                return cls.dictionary
            try:
                json.dumps(obj)
            except (TypeError, OverflowError):
                return cls.pkl
            return cls.dictionary
        elif isinstance(obj, list) or obj is list:
            if obj is list:
                return cls.array
            try:
                json.dumps(obj)
            except (TypeError, OverflowError):
                return cls.pkl
            return cls.array
        elif isinstance(obj, io.BufferedIOBase) or obj is File or isinstance(obj, File):
            return cls.file
        else:
            return cls.pkl

    @staticmethod
    def to_object(io_type: "RunIOType"):
        if isinstance(io_type, str):
            io_type = RunIOType(io_type)

        if io_type == RunIOType.integer:
            return int
        elif io_type == RunIOType.fp:
            return float
        elif io_type == RunIOType.string:
            return str
        elif io_type == RunIOType.boolean:
            return bool
        elif io_type == RunIOType.none:
            return None
        elif io_type == RunIOType.dictionary:
            return dict
        elif io_type == RunIOType.array:
            return list
        else:
            raise ValueError(f"Invalid io_type: {io_type}")


class RunOutputFile(BaseModel):
    name: str
    path: str
    url: t.Optional[str]
    size: int


class RunOutput(BaseModel):
    type: RunIOType
    value: t.Optional[t.Any]
    file: t.Optional[RunOutputFile]


class RunResult(BaseModel):
    run_id: str
    outputs: t.List[RunOutput]

    def result_array(self) -> t.List[t.Any]:
        # return [output.value for output in self.outputs]
        output_array = []
        for output in self.outputs:
            if output.type == RunIOType.file:
                # output_array.append(output.value)
                from pipeline.objects.graph import File

                if output.file is not None and output.file.url is not None:
                    output_array.append(File(url=output.file.url))
                else:
                    raise Exception("Returned file missing information.")
            else:
                output_array.append(output.value)
        return output_array


class RunInput(BaseModel):
    type: RunIOType
    value: t.Any

    file_name: t.Optional[str]
    file_path: t.Optional[str]
    # The file URL is only populated when this schema is
    # returned by the API, the user should never populate it
    file_url: t.Optional[str]


class ContainerRunErrorType(str, Enum):
    input_error = "input_error"
    cuda_oom = "cuda_oom"
    cuda_error = "cuda_error"
    oom = "oom"
    pipeline_error = "pipeline_error"
    startup_error = "startup_error"
    pipeline_loading = "pipeline_loading"
    unknown = "unknown"


class ContainerRunError(BaseModel):
    type: ContainerRunErrorType
    message: str
    traceback: t.Optional[str]


class ContainerRunCreate(BaseModel):
    # run_id is optional since it's just used for attaching logs to a run
    run_id: t.Optional[str]
    inputs: t.List[RunInput]


class ContainerRunResult(BaseModel):
    inputs: t.Optional[t.List[RunInput]]
    outputs: t.Optional[t.List[RunOutput]]
    error: t.Optional[ContainerRunError]

    def outputs_formatted(self) -> t.List[t.Any]:
        outputs = self.outputs or []
        output_array = []
        for output in outputs:
            if output.type == RunIOType.file:
                from pipeline.objects.graph import File

                if output.file is not None and output.file.url is not None:
                    output_array.append(File(url=output.file.url))
                else:
                    raise Exception("Returned file missing information.")
            else:
                output_array.append(output.value)
        return output_array


class ClusterRunResult(ContainerRunResult):
    id: str

    created_at: datetime
    updated_at: datetime

    pipeline_id: str

    state: RunState

    class Config:
        orm_mode = True


class RunStateTransition(BaseModel):
    run_id: str
    new_state: RunState
    time: datetime


class RunStateTransitions(BaseModel):
    """View for all state transitions of a given run"""

    data: t.List[RunStateTransition]


class RunCreate(ContainerRunCreate):
    # pipeline id or pointer
    pipeline: str
    async_run: bool = False



from pipeline.objects.function import Function
from pipeline.objects.graph import Graph, GraphNode, Variable


class Pipeline:
    _current_pipeline: Graph
    _pipeline_context_active: bool = False

    def __init__(
        self,
    ):
        ...

    def __enter__(self):
        Pipeline._pipeline_context_active = True

        Pipeline._current_pipeline = Graph()

        return self

    def __exit__(self, type, value, traceback):
        Pipeline._pipeline_context_active = False

    def output(self, *outputs: Variable) -> None:
        for _output in outputs:
            variable_index = Pipeline._current_pipeline.variables.index(_output)
            if variable_index != -1:
                Pipeline._current_pipeline.variables[variable_index].is_output = True

                Pipeline._current_pipeline.outputs.append(
                    Pipeline._current_pipeline.variables[variable_index]
                )
                for variable in Pipeline._current_pipeline.variables:
                    if variable.local_id == _output.local_id:
                        variable.is_output = True
                        break

    def get_pipeline(self) -> Graph:
        return Pipeline._current_pipeline

    @staticmethod
    def add_function(function: Function) -> None:
        if Pipeline._pipeline_context_active:
            Pipeline._current_pipeline.functions.append(function)
        else:
            raise Exception("Cant add a function when not defining a pipeline!")

    @staticmethod
    def add_graph_node(graph_node: GraphNode) -> None:
        if Pipeline._pipeline_context_active:
            Pipeline._current_pipeline.nodes.append(graph_node)
        else:
            raise Exception("Cant add a node when not defining a pipeline!")



import hashlib
import importlib
import os
import traceback
import typing as t
import urllib.parse
from pathlib import Path
from types import NoneType, UnionType
from urllib import request
from urllib.parse import urlparse

from loguru import logger

from pipeline.cloud.schemas import pipelines as pipeline_schemas
from pipeline.cloud.schemas import runs as run_schemas
from pipeline.exceptions import RunInputException, RunnableError
from pipeline.objects import Directory, File, Graph
from pipeline.objects.graph import InputSchema


def is_url(string):
    try:
        result = urlparse(string)
        return all([result.scheme, result.netloc])
    except ValueError:
        return False


def _get_url_or_path(input_schema: run_schemas.RunInput) -> str | None:
    return input_schema.file_url if input_schema.file_url else input_schema.file_path


class Manager:
    def __init__(self, pipeline_path: str):
        if ":" not in pipeline_path:
            raise ValueError(
                "Invalid pipeline path, must be in format <module_or_file>:<pipeline>"
            )
        if len(pipeline_path.split(":")) != 2:
            raise ValueError(
                "Invalid pipeline path, must be in format <module_or_file>:<pipeline>"
            )

        self.pipeline_path = pipeline_path
        self.pipeline_module_str, self.pipeline_name_str = pipeline_path.split(":")

        self.pipeline_state: pipeline_schemas.PipelineState = (
            pipeline_schemas.PipelineState.not_loaded
        )
        self.pipeline_state_message: str | None = None

        try:
            self.pipeline_module = importlib.import_module(self.pipeline_module_str)
            self.pipeline: Graph = getattr(self.pipeline_module, self.pipeline_name_str)
        except ModuleNotFoundError:
            raise ValueError(f"Could not find module {self.pipeline_module_str}")
        except AttributeError:
            raise ValueError(
                (
                    f"Could not find pipeline {self.pipeline_name_str} in module"
                    f" {self.pipeline_module_str}"
                )
            )
        except Exception as e:
            raise ValueError(f"Unexpected error: {e}")

        self.pipeline_name = os.environ.get("PIPELINE_NAME", "unknown")
        self.pipeline_image = os.environ.get("PIPELINE_IMAGE", "unknown")

        logger.info(f"Pipeline set to {self.pipeline_path}")

    def startup(self):
        # add context to enable fetching of startup logs
        with logger.contextualize(pipeline_stage="startup"):
            logger.info("Starting pipeline")
            self.pipeline_state = pipeline_schemas.PipelineState.loading
            try:
                self.pipeline._startup()
            except Exception:
                tb = traceback.format_exc()
                logger.exception("Exception raised during pipeline execution")
                self.pipeline_state = pipeline_schemas.PipelineState.failed
                self.pipeline_state_message = tb
            else:
                self.pipeline_state = pipeline_schemas.PipelineState.loaded
                logger.info("Pipeline started successfully")

    def _resolve_file_variable_to_local(
        self,
        file: File | Directory,
        *,
        use_tmp: bool = False,
    ) -> None:
        local_host_dir = "/tmp"

        if hasattr(file, "url") and file.url is not None:
            cache_name = hashlib.md5(file.url.geturl().encode()).hexdigest()

            file_name = file.url.geturl().split("/")[-1]
            local_path = f"{local_host_dir}/{cache_name}/{file_name}"
            file_path = Path(local_path)
            file_path.parent.mkdir(parents=True, exist_ok=True)

            request.urlretrieve(file.url.geturl(), local_path)
        elif file.remote_id is not None or file.path is not None:
            local_path = Path(file.path)
            local_path.parent.mkdir(parents=True, exist_ok=True)
        else:
            raise Exception("File not found, must pass in URL, Path, or Remote ID.")

        if isinstance(file, Directory):
            raise NotImplementedError("Remote ID not implemented yet")

            # file.path = Path(f"{local_host_dir}/{cache_name}_dir")

            # with zipfile.ZipFile(local_path, "r") as zip_ref:
            #     zip_ref.extractall(str(file.path))
            # return

        file.path = Path(local_path)

    def _create_file_variable(
        self,
        *,
        path_or_url: str | None = None,
        is_directory: bool = False,
        use_tmp: bool = False,
    ) -> File | Directory:
        path: str | None = None
        url: str | None = None

        if is_url(path_or_url):
            url = str(urllib.parse.urlparse(path_or_url).geturl())
        else:
            path = path_or_url

        if path is None and url is None:
            raise Exception("Must provide either path or url")

        variable: File | Directory | None = None

        if is_directory:
            variable = Directory(path=path, url=url)
        else:
            variable = File(path=path, url=url)

        self._resolve_file_variable_to_local(variable, use_tmp=use_tmp)
        return variable

    def _parse_inputs(
        self, input_data: t.List[run_schemas.RunInput] | None, graph: Graph
    ) -> t.List[t.Any]:
        inputs = []
        if input_data is None:
            input_data = []
            return input_data
        if len(input_data) > 0:
            for item in input_data:
                input_schema = run_schemas.RunInput.parse_obj(item)
                if input_schema.type == run_schemas.RunIOType.file:
                    if input_schema.file_path is None and input_schema.file_url is None:
                        raise RunInputException(
                            "A file must either have a path or url attribute"
                        )
                    path_or_url = _get_url_or_path(input_schema)
                    if path_or_url is None:
                        raise RunInputException(
                            "A file must either have a path or url attribute"
                        )

                    variable = self._create_file_variable(path_or_url=path_or_url)
                    inputs.append(
                        variable,
                    )

                else:
                    inputs.append(input_schema.value)
        graph_inputs = list(filter(lambda v: v.is_input, graph.variables))

        if len(inputs) != len(graph_inputs):
            raise RunInputException("Inputs do not match graph inputs")

        final_inputs = []

        for i, (user_input, model_input) in enumerate(zip(inputs, graph_inputs)):
            target_type = model_input.type_class
            if issubclass(target_type, InputSchema) and isinstance(user_input, dict):
                # schema_instance = target_type()
                annotations = target_type.__annotations__
                for key, value in annotations.items():
                    if isinstance(value, UnionType) or "typing.Optional" in str(value):
                        var_union_types = list(t.get_args(value))
                        if len(var_union_types) > 2:
                            raise RunInputException("Only support Union of 2 types")
                        if NoneType in var_union_types:
                            var_union_types.remove(NoneType)
                        else:
                            raise RunInputException("Only support Union with None")
                        var_type = var_union_types[0]
                    else:
                        var_type = value

                    if var_type == File:
                        if user_input.get(key) is None:
                            continue
                        file_schema = run_schemas.RunInput.parse_obj(user_input[key])
                        path_or_url = _get_url_or_path(file_schema)
                        variable = self._create_file_variable(
                            path_or_url=path_or_url,
                            use_tmp=True,
                        )
                        user_input[key] = variable

            final_inputs.append(user_input)
        return final_inputs

    def run(
        self, run_id: str | None, input_data: t.List[run_schemas.RunInput] | None
    ) -> t.Any:
        with logger.contextualize(run_id=run_id):
            logger.info("Running pipeline")
            args = self._parse_inputs(input_data, self.pipeline)
            try:
                result = self.pipeline.run(*args)
            except Exception as exc:
                raise RunnableError(exception=exc, traceback=traceback.format_exc())
            return result



from pipeline.configuration import current_configuration
from pipeline.objects import File, Pipeline, Variable, entity, onnx_to_pipeline, pipe

__all__ = [
    "Pipeline",
    "Variable",
    "entity",
    "pipe",
    "File",
    "onnx_to_pipeline",
    "current_configuration",
]



import io
import os
import uuid
from pathlib import Path
from zipfile import ZipFile

import httpx
from httpx import HTTPStatusError
from tqdm import tqdm
from tqdm.utils import CallbackIOWrapper

from pipeline.cloud import http
from pipeline.cloud.schemas import files as s
from pipeline.cloud.schemas.runs import RunInput, RunIOType
from pipeline.objects import Directory, File
from pipeline.util import CallbackBytesIO
from pipeline.util.logging import PIPELINE_FILE_STR

FILE_CHUNK_SIZE = 200 * 1024 * 1024  # 200 MiB


def upload_multipart_file(file_path: Path, progress: bool = False) -> s.FileGet:
    """Upload file directly to storage using multi-part upload.

    Since files can be very large (e.g. model weights), we use the following process:
    - We need to split the file into chunks based on FILE_CHUNK_SIZE
    - We first initialise the multi-part upload with the server
    - We then upload the file in chunks (requesting a presigned upload URL for each
        chunk beforehand)
    - Lastly, we finalise the multi-part upload with the server
    """

    file_size = os.path.getsize(file_path)
    init_result = _init_multipart_upload(file_path.name)
    file_id = init_result.file_id
    upload_id = init_result.upload_id

    progress_bar = None
    if progress:
        progress_bar = tqdm(
            desc=f"{PIPELINE_FILE_STR} Uploading {file_path}",
            unit="B",
            unit_scale=True,
            total=file_size,
            unit_divisor=1024,
        )
    parts = []
    with file_path.open("rb") as f:
        while True:
            file_data = f.read(FILE_CHUNK_SIZE)
            if not file_data:
                if progress_bar:
                    progress_bar.close()
                break
            part_num = len(parts) + 1
            # If displaying progress bar then wrap our data object in a tqdm callback
            if progress_bar:
                data = CallbackBytesIO(progress_bar.update, file_data)
            else:
                data = io.BytesIO(file_data)

            upload_metadata = _upload_multipart_file_chunk(
                data=data,
                file_id=file_id,
                upload_id=upload_id,
                part_num=part_num,
            )
            parts.append(upload_metadata)

    file_get = _finalise_multipart_upload(
        file_id=file_id, upload_id=upload_id, multipart_metadata=parts
    )
    return file_get


def _init_multipart_upload(filename: str) -> s.MultipartFileUploadInitGet:
    res = http.post(
        "/v3/pipeline_files/initiate-multipart-upload",
        s.MultipartFileUploadInitCreate(name=filename).dict(),
    )
    return s.MultipartFileUploadInitGet.parse_obj(res.json())


def _upload_multipart_file_chunk(
    data: io.BytesIO | CallbackIOWrapper,
    file_id: str,
    upload_id: str,
    part_num: int,
) -> s.MultipartFileUploadMetadata:
    """Upload a single chunk of a multi-part pipeline file upload.

    Returns the metadata associated with this upload (this is needed to pass into
    the finalisation step).
    """
    # get presigned URL
    part_upload_schema = s.MultipartFileUploadPartCreate(
        file_id=file_id,
        upload_id=upload_id,
        part_num=part_num,
    )

    res = http.post(
        "/v3/pipeline_files/presigned-url",
        part_upload_schema.dict(),
    )

    part_upload_get = s.MultipartFileUploadPartGet.parse_obj(res.json())

    response = httpx.put(
        part_upload_get.upload_url,
        content=data,
        timeout=60,
    )
    response.raise_for_status()
    etag = response.headers["ETag"]
    return s.MultipartFileUploadMetadata(ETag=etag, PartNumber=part_num)


def _finalise_multipart_upload(
    file_id: str,
    upload_id: str,
    multipart_metadata: list[s.MultipartFileUploadMetadata],
) -> s.FileGet:
    """Finalise the direct multi-part file upload"""
    finalise_upload_schema = s.MultipartFileUploadFinaliseCreate(
        file_id=file_id,
        upload_id=upload_id,
        multipart_metadata=multipart_metadata,
    )
    response = http.post(
        "/v3/pipeline_files/finalise-multipart-upload",
        json_data=finalise_upload_schema.dict(),
    )
    return s.FileGet.parse_obj(response.json())


def get_path_from_id(file_id: str) -> str:
    file_get_response = http.get(f"/v3/pipeline_files/{file_id}")
    if file_get_response is None:
        raise ValueError(f"Received no response for file_id {file_id}")

    file_get_schema = s.FileGet.parse_obj(file_get_response.json())

    return file_get_schema.path


def create_remote_directory(local_path: Path) -> s.FileGet:
    local_path_str = str(local_path)
    if not local_path.exists():
        raise FileNotFoundError(f"Directory not found (path={local_path_str}) ")
    if not local_path.is_dir() and not local_path_str.endswith(".zip"):
        raise ValueError(f"Path is not a directory or zip file (path={local_path_str})")

    zip_path = local_path

    if not local_path_str.endswith(".zip"):
        tmp_path = Path("/tmp") / (str(uuid.uuid4()) + ".zip")
        with ZipFile(str(tmp_path), "w") as zip_file:
            for root, dirs, files in os.walk(local_path_str):
                for file in files:
                    zip_file.write(
                        os.path.join(root, file),
                        arcname=file,
                    )
        zip_path = tmp_path

    try:
        file_get = upload_multipart_file(zip_path, progress=True)
    except HTTPStatusError as e:
        if e.response.status_code == 403:
            raise Exception(
                f"Permission denied uploading directory (path={local_path_str})"
            )
        raise Exception(f"Error uploading directory (path={local_path_str}): {e}")

    return file_get


def resolve_pipeline_file_object(obj: File | Directory) -> None:
    # Handle from ID, URL, or local path
    # Either URL or path has to be popluated, and on the remote

    if obj.url is not None:
        return
    elif obj.remote_id is not None:
        obj.path = Path(get_path_from_id(obj.remote_id))
        return
    if obj.path:
        if isinstance(obj, Directory):
            remote_dir = create_remote_directory(obj.path)
            obj.path = Path(remote_dir.path)
            return
        elif isinstance(obj, File):
            remote_file = upload_multipart_file(obj.path)
            obj.path = Path(remote_file.path)
            return


def resolve_run_input_file_object(obj: File | Directory) -> RunInput:
    if obj.url is not None:
        return RunInput(
            type=RunIOType.file,
            value=None,
            file_name=obj.url.geturl().split("/")[-1],
            file_path=obj.url.geturl(),
        )
    elif obj.remote_id is not None:
        path = get_path_from_id(obj.remote_id)
        return RunInput(
            type=RunIOType.file,
            value=None,
            file_name=path.split("/")[-1],
            file_path=path,
        )
    elif obj.path is not None:
        if isinstance(obj, Directory):
            remote_dir = create_remote_directory(obj.path)
            return RunInput(
                type=RunIOType.file,
                value=None,
                file_name=remote_dir.path.split("/")[-1],
                file_path=remote_dir.path,
            )
        elif isinstance(obj, File):
            remote_file = upload_multipart_file(obj.path)
            return RunInput(
                type=RunIOType.file,
                value=None,
                file_name=remote_file.path.split("/")[-1],
                file_path=remote_file.path,
            )

    raise Exception(f"Invalid file object: {obj}, must have remote_id, path, or URL")



import typing as t

import httpx

from pipeline.cloud import http
from pipeline.cloud.pipelines import _data_to_run_input
from pipeline.cloud.schemas.runs import Run, RunCreate, RunState


async def run_pipeline(
    pipeline: str,
    *data,
    async_run: bool = False,
    return_response: bool = False,
) -> t.Union[Run, httpx.Response]:
    run_create_schema = RunCreate(
        pipeline=pipeline,
        inputs=_data_to_run_input(data),
        async_run=async_run,
    )

    res = await http.async_post(
        "/v4/runs",
        json_data=run_create_schema.dict(),
        handle_error=not return_response,
    )

    if return_response:
        return res

    run_get = Run.parse_obj(res.json())

    if RunState.is_terminal(run_get.state) and run_get.state != RunState.completed:
        raise Exception(
            f"Remote Run '{run_get.id}' failed with exception: "
            f"{getattr(getattr(run_get, 'error', None), 'traceback', None)}"
        )

    return run_get



import functools
import io
import os
import typing as t
from json.decoder import JSONDecodeError

import httpx
import requests
from requests_toolbelt import MultipartEncoder, MultipartEncoderMonitor
from tqdm import tqdm

from pipeline import current_configuration
from pipeline.util.logging import PIPELINE_STR, _print

_client = None
_client_async = None


class APIError(Exception):
    def __init__(
        self, url: str, status_code: int, detail: dict | str, request_id: str | None
    ):
        self.url = url
        self.status_code = status_code
        self.detail = detail
        self.request_id = request_id
        super().__init__(url, status_code, detail, request_id)

    @classmethod
    def from_response(cls, response: httpx.Response):
        try:
            detail = response.json()["detail"]
            if not isinstance(detail, dict):
                detail = {"detail": detail}
        except (JSONDecodeError, TypeError, KeyError):
            detail = {
                "message": "Something went wrong.",
                "response": response.text,
            }
        return cls(
            url=str(response.url),
            detail=detail,
            status_code=response.status_code,
            request_id=response.headers.get("x-correlation-id"),
        )

    def __str__(self):
        error_msg = ", ".join(f"{k}={v}" for k, v in self.__dict__.items())
        error_msg = f"APIError({error_msg})"
        return error_msg


def raise_if_http_status_error(response: httpx.Response):
    try:
        response.raise_for_status()
    except httpx.HTTPStatusError as e:
        error = APIError.from_response(e.response)
        _print(str(error), level="ERROR")
        raise error from None


def handle_http_status_error(func):
    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        response = func(*args, **kwargs)
        if kwargs.pop("handle_error", True):
            raise_if_http_status_error(response)
        return response

    return wrapper


def _get_client() -> httpx.Client:
    global _client
    if _client is None:
        url = (
            active_remote.url
            if (active_remote := current_configuration.active_remote) is not None
            else os.environ.get("PIPELINE_API_URL", "https://www.mystic.ai/")
        )
        token = (
            current_configuration.active_remote.token
            if current_configuration.active_remote is not None
            else os.environ.get("PIPELINE_API_TOKEN", None)
        )
        _client = httpx.Client(
            base_url=url,
            headers={
                "Authorization": f"Bearer {token}",
            },
            timeout=300,
        )
    return _client


def _get_async_client() -> httpx.AsyncClient:
    global _client_async
    if _client_async is None:
        url = (
            active_remote.url
            if (active_remote := current_configuration.active_remote) is not None
            else os.environ.get("PIPELINE_API_URL", "https://www.mystic.ai/")
        )
        token = (
            current_configuration.active_remote.token
            if current_configuration.active_remote is not None
            else os.environ.get("PIPELINE_API_TOKEN", None)
        )
        _client_async = httpx.AsyncClient(
            base_url=url,
            headers={
                "Authorization": f"Bearer {token}",
            },
            timeout=300,
        )
    return _client_async


@handle_http_status_error
def post(
    endpoint: str,
    json_data: dict = None,
    handle_error: bool = True,
) -> httpx.Response:
    client = _get_client()
    response = client.post(endpoint, json=json_data)
    return response


@handle_http_status_error
async def async_post(
    endpoint: str,
    json_data: dict = None,
    handle_error: bool = True,
) -> httpx.Response:
    client = _get_async_client()
    response = await client.post(
        endpoint,
        json=json_data,
    )
    return response


@handle_http_status_error
def patch(
    endpoint: str,
    json_data: dict = None,
    handle_error: bool = True,
) -> httpx.Response:
    client = _get_client()
    return client.patch(
        endpoint,
        json=json_data,
    )


@handle_http_status_error
def get(
    endpoint: str,
    handle_error: bool = True,
    **kwargs,
) -> httpx.Response:
    client = _get_client()
    response = client.get(endpoint, **kwargs)
    return response


@handle_http_status_error
def delete(
    endpoint: str,
    handle_error: bool = True,
    **kwargs,
) -> httpx.Response:
    client = _get_client()
    response = client.delete(endpoint, **kwargs)
    return response


def create_callback(encoder: MultipartEncoder) -> t.Callable:
    encoder_len = encoder.len
    bar = tqdm(
        desc=f"{PIPELINE_STR} Uploading",
        unit="B",
        unit_scale=True,
        total=encoder_len,
        unit_divisor=1024,
    )

    def callback(monitor):
        bar.n = monitor.bytes_read
        bar.refresh()
        if monitor.bytes_read == encoder_len:
            bar.close()

    return callback


def get_progress_bar_uploader(
    files: t.Dict[str, io.BufferedIOBase], params: t.Dict[str, str]
) -> MultipartEncoderMonitor:
    encoder = create_upload(files, params)
    callback = create_callback(encoder)
    monitor = MultipartEncoderMonitor(encoder, callback)
    return monitor


def create_upload(
    files: t.Dict[str, io.BufferedIOBase], params: t.Dict[str, str]
) -> MultipartEncoder:
    file_dict = {
        param_key: (
            getattr(file, "name", "undefined"),
            file,
            "application/octet-stream",
            {
                "Content-Transfer-Encoding": "binary",
            },
        )
        for param_key, file in files.items()
    }

    return MultipartEncoder(
        dict(
            **file_dict,
            **params,
        ),
    )


def post_files(
    endpoint: str,
    files: t.Dict[str, t.Any],
    params: t.Dict[str, t.Any] = dict(),
    data: t.Dict[str, t.Any] = dict(),
    progress: bool = False,
) -> httpx.Response:
    if progress:
        monitor = get_progress_bar_uploader(files=files, params=params)
        url = (
            active_remote.url
            if (active_remote := current_configuration.active_remote) is not None
            else os.environ.get("PIPELINE_API_URL", "https://www.mystic.ai/")
        )
        token = (
            current_configuration.active_remote.token
            if current_configuration.active_remote is not None
            else os.environ.get("PIPELINE_API_TOKEN", None)
        )
        response = requests.post(
            f"{url}{endpoint}",
            data=monitor,
            headers={
                "content-Type": monitor.content_type,
                "Authorization": f"Bearer {token}",
            },
            params=params,
        )
    else:
        client = _get_client()
        response = client.post(
            endpoint,
            files=files,
            data=data,
            params=params,
        )

    return response
